/*
  Project Name: Brain
  Description: This program is the central hub of the entire table control system. 
  On one hand, it acts as the master for ESP-NOW, coordinating communication 
  between the website, slaves, and the logic. On the other hand, it hosts the 
  website for the game master and communicates with it.
  Author: Lena Hagenauer
  Date: 05 January 2025
  Version: 2.0
  License: MIT License
*/

// Include libraries
#include <FS.h>
#include <LittleFS.h>
#include <esp_now.h>
#include <WiFi.h>
#include <esp_wifi.h>
#include "ESPAsyncWebServer.h"
#include "AsyncTCP.h"
#include <ArduinoJson.h>
#include <List.hpp>

// Adressen 0 = server; 1 = Licht; 2 = Spielereinfolge; 4 = WebPage; 5 = offen; 6+ Spieler

// Replace with your network credentials (STATION)
//const char* ssid = "YourSSID";
//const char* password = "YourPASSWORD";


esp_now_peer_info_t slave;
int chan; 

// Message types
enum MessageType {PAIRING, DATA, LIGHT, INITATIVE};
MessageType messageType;

// variables definitions
int counter = 0;
int initativ = 0;

uint8_t clientMacAddress[6];// Client MAC address
uint8_t lightMacAddress[6];// Light MAC address
uint8_t initativeMacAddress[6];// Initiative MAC address

// Structure to text message
typedef struct struct_message {       //new Structur for message
  uint8_t msgType;
  uint8_t id; //board ID
  char message[168];
} struct_message;

// Structure for pairing message
typedef struct struct_pairing {       // new structure for pairing
    uint8_t msgType;
    uint8_t id; //board ID
    uint8_t macAddr[6];
    uint8_t channel;
    char name[64];
} struct_pairing;

// Structure to send light Information
typedef struct struct_light {
  uint8_t msgType;
  char colorType[8];
  uint32_t color;
} struct_light;

// Structure to send Informtion to Initative
typedef struct struct_initative {
  uint8_t msgType;
  int nr;
  int state; //add 1; sup 2; end 99;
} struct_initative;

// Variables definitions
struct_message incomingReadings;
struct_message outgoingSetpoints;
struct_pairing pairingData;
struct_light outputLight;
struct_initative outputInitative;

// Initializes an asynchronous web server
AsyncWebServer server(80);
AsyncEventSource events("/events");

// Init empty player lsit
List<struct struct_pairing> playerList;

// Function to read the MAC address
void readMacAddress(){
  uint8_t baseMac[6];
  esp_err_t ret = esp_wifi_get_mac(WIFI_IF_STA, baseMac);
  if (ret == ESP_OK) {
    Serial.printf("%02x:%02x:%02x:%02x:%02x:%02x\n",
                  baseMac[0], baseMac[1], baseMac[2],
                  baseMac[3], baseMac[4], baseMac[5]);
  } else {
    Serial.println("Failed to read MAC address");
  }
}

// Function to print the MAC address
void printMAC(const uint8_t * mac_addr){
  char macStr[18];
  snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
           mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5]);
  Serial.print(macStr);
}

// Function to add a peer
bool addPeer(const uint8_t *peer_addr) {      // add pairing
  memset(&slave, 0, sizeof(slave));
  const esp_now_peer_info_t *peer = &slave;
  memcpy(slave.peer_addr, peer_addr, 6);
  
  slave.channel = chan; // pick a channel
  slave.encrypt = 0; // no encryption
  // check if the peer exists
  bool exists = esp_now_is_peer_exist(slave.peer_addr);
  if (exists) {
    // Slave already paired.
    Serial.println("Already Paired");
    return false;//true
  }
  else {
    esp_err_t addStatus = esp_now_add_peer(peer);
    if (addStatus == ESP_OK) {
      // Pair success
      Serial.println("Pair success");
      return true;
    }
    else 
    {
      Serial.println("Pair failed");
      return false;
    }
  }
} 

// Callback when data is sentt
void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  Serial.print("Last Packet Send Status: ");
  Serial.print(status == ESP_NOW_SEND_SUCCESS ? "Delivery Success to " : "Delivery Fail to ");
  printMAC(mac_addr);
  Serial.println();
}


void OnDataRecv(const uint8_t * mac_addr, const uint8_t *incomingData, int len) { 
 
  Serial.print(len);
  Serial.println(" bytes of new data received.");
  StaticJsonDocument<1000> root;
  StaticJsonDocument<1000> player;
  String payload;
  uint8_t type = incomingData[0];// first message byte is the type of message 
  switch (type) {
  case DATA :// Incoming Mesage: send Player name to Webside
    memcpy(&incomingReadings, incomingData, sizeof(incomingReadings));
    // create a JSON document with received data and send it by event to the web page
    root["id"] = incomingReadings.id;
    root["message"] = (String) incomingReadings.message;
    serializeJson(root, payload);
    Serial.print("event send :");
    serializeJson(root, Serial);
    events.send(payload.c_str(), "new_readings", millis());
    Serial.println();
    break;
  
  case PAIRING:// Incoming pairing: Pairs a new slave depending on type
    memcpy(&pairingData, incomingData, sizeof(pairingData));
    struct_pairing playerData;
    memcpy(&playerData, incomingData, sizeof(playerData));
    Serial.print("Pairing request from MAC Address: ");
    printMAC(pairingData.macAddr);
    Serial.print(" on channel ");
    Serial.println(pairingData.channel);

    clientMacAddress[0] = pairingData.macAddr[0];
    clientMacAddress[1] = pairingData.macAddr[1];
    clientMacAddress[2] = pairingData.macAddr[2];
    clientMacAddress[3] = pairingData.macAddr[3];
    clientMacAddress[4] = pairingData.macAddr[4];
    clientMacAddress[5] = pairingData.macAddr[5];

    if (pairingData.id > 0) {     // do not replay to server itself
      if (pairingData.msgType == PAIRING) { 
        player["id"] = pairingData.id;
        player["name"] = (String) pairingData.name;
        int idClient = pairingData.id;
        pairingData.id = 0;       // 0 is server
        // Server is in AP_STA mode: peers need to send data to server soft AP MAC address 
        WiFi.softAPmacAddress(pairingData.macAddr);
        Serial.print("Pairing MAC Address: ");
        printMAC(clientMacAddress);
        pairingData.channel = chan;
        Serial.println(" send response");
        esp_err_t result = esp_now_send(clientMacAddress, (uint8_t *) &pairingData, sizeof(pairingData));
        if(addPeer(clientMacAddress)&&(idClient > 5)){ //add player if not exist
  
          //add play to list
          playerList.add((struct struct_pairing) playerData);

          // create a JSON document with received data and send it by event to the web page
          serializeJson(player, payload);
          Serial.print("event send :");
          serializeJson(player, Serial);
          events.send(payload.c_str(), "player_login", millis());
          Serial.println(); 

        }else if(idClient == 1){// add light
          lightMacAddress[0] = playerData.macAddr[0];
          lightMacAddress[1] = playerData.macAddr[1];
          lightMacAddress[2] = playerData.macAddr[2];
          lightMacAddress[3] = playerData.macAddr[3];
          lightMacAddress[4] = playerData.macAddr[4];
          lightMacAddress[5] = playerData.macAddr[5];
          Serial.println("add light"); 
          //printMAC(lightMacAddress);
        }else if(idClient == 2){// add initative
          initativeMacAddress[0] = playerData.macAddr[0];
          initativeMacAddress[1] = playerData.macAddr[1];
          initativeMacAddress[2] = playerData.macAddr[2];
          initativeMacAddress[3] = playerData.macAddr[3];
          initativeMacAddress[4] = playerData.macAddr[4];
          initativeMacAddress[5] = playerData.macAddr[5];
          Serial.println("add initative"); 

        }
      }  
    }  
    break; 
  }
}

// Init ESP-NOW
void initESP_NOW(){
    if (esp_now_init() != ESP_OK) {
      Serial.println("Error initializing ESP-NOW");
      return;
    }
    esp_now_register_send_cb(OnDataSent);
    esp_now_register_recv_cb(esp_now_recv_cb_t(OnDataRecv));
} 

void setup() {
  // Initialize Serial Monitor
  Serial.begin(115200);

  // Initialize LittleFS
  if(!LittleFS.begin()){
    Serial.println("An Error has occurred while mounting LittleFS");
    return;
  }else{
    Serial.println("LittleFS mounted successfully");
  }

  // Set Wi-Fi mode to Station
  WiFi.mode(WIFI_STA);
  WiFi.STA.begin();
  Serial.print("Server MAC Address: ");
  readMacAddress();

  // Set the device as a Station and Soft Access Point simultaneously
  WiFi.mode(WIFI_AP_STA);
  // Set device as a Wi-Fi Station
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Setting as a Wi-Fi Station..");
  }

  Serial.print("Server SOFT AP MAC Address:  ");
  Serial.println(WiFi.softAPmacAddress());

  chan = WiFi.channel();
  Serial.print("Station IP Address: ");
  Serial.println(WiFi.localIP());
  Serial.print("Wi-Fi Channel: ");
  Serial.println(WiFi.channel());

  // Initialize ESP-NOW
  initESP_NOW();

  // Serve the index.html file
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send( LittleFS, "/index.html", "text/html");
  });
  // Route to load chat.css file
  server.on("/css/chat.css", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(LittleFS, "/css/chat.css", "text/css");
  });
  // Route to load color.css file
  server.on("/css/color.css", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(LittleFS, "/css/color.css", "text/css");
  });
  // Route to load general.css file
  server.on("/css/general.css", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(LittleFS, "/css/general.css", "text/css");
  });
  // Route to load rules.css file
  server.on("/css/rules.css", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(LittleFS, "/css/rules.css", "text/css");
  });
  // Route to load initativ.css file
  server.on("/css/initativ.css", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(LittleFS, "/css/initativ.css", "text/css");
  });
  // Route to load scipt.js file
  server.on("/scripts/script.js", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(LittleFS, "/scripts/script.js", "text/js");
  });
  // Route to load symbol.png
  server.on("/src/symbol.png", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(LittleFS, "/src/symbol.png", "text/js");
  });


  // Handle events
  events.onConnect([](AsyncEventSourceClient *client){
    if(client->lastId()){
      Serial.printf("Client reconnected! Last message ID that it got is: %u\n", client->lastId());
    }
    // send event with message "hello!", id current millis
    // and set reconnect delay to 1 second
    client->send("hello!", NULL, millis(), 10000);
  });
  server.addHandler(&events);
  //incoming Messages

  //read Color from Colorpicker
  server.on("/color", HTTP_GET, [](AsyncWebServerRequest *request){
    String colorState;
    uint32_t color;
    
    uint8_t playerAddressX[6];

    if (request->hasParam("type")) {
      colorState = request->getParam("type")->value();
      
    }
    String colorString;
    if (request->hasParam("color")) {
      colorString = request->getParam("color")->value();
      char colorChar[8];
      for(int i = 0; i<8;i++){ 
        colorChar[i]=colorString[i+1];
      }
      color = strtoul(colorChar, NULL, 16);
    }

    Serial.println(colorString);
    Serial.println((String)color);
    
    outputLight.msgType = LIGHT;
    colorState.toCharArray(outputLight.colorType, 8);
    outputLight.color = color;

    Serial.println((String)outputLight.color);

    esp_now_send(lightMacAddress, (uint8_t *) &outputLight, sizeof(outputLight));
    
    request->send(LittleFS, "/index.html", String(), false);
  });

//read initative and save current position in variable initativ, starting at 0
  server.on("/initative", HTTP_GET, [](AsyncWebServerRequest *request){
      String initativeCount;
     
      int initativ = 0;
      int stateInitaive = 0;
      if (request->hasParam("initative")) {
        initativeCount = request->getParam("initative")->value();
        if(initativeCount == "add") {
          //ToDo: add 1 to numbers of entities in initative
          stateInitaive = 1;
        } else if(initativeCount == "sub") {
          //ToDo: substract 1 from numbers of entities in initative
          stateInitaive = 2;
        } else if (initativeCount == "end") {
          stateInitaive = 99;
        } else {
          //ToDo: set number of players in initative to 
          // initativeCount.toInt();
          // and start initative
          initativ = initativeCount.toInt();
        }
      } 
       
      outputInitative.msgType = INITATIVE;
      outputInitative.nr = initativ;
      outputInitative.state = stateInitaive; //add 1; sup 2; end 99;

      Serial.print("Initative: ");
      Serial.println(initativ);
       Serial.print("stateInitaive: ");
      Serial.println(stateInitaive);

      esp_now_send(initativeMacAddress, (uint8_t *) &outputInitative, sizeof(outputInitative));

      request->send(LittleFS, "/index.html", String(), false);
    });

  // Read text imput
  server.on("/get", HTTP_GET, [] (AsyncWebServerRequest *request) {
    String inputMessage;
    struct_pairing currentPlayer;    
    uint8_t playerAddressX[6];
    bool all = true;

    if (request->hasParam("message")) {
      inputMessage = request->getParam("message")->value();
    }

    outgoingSetpoints.msgType = DATA;
    outgoingSetpoints.id = 0;
    strcpy(outgoingSetpoints.message, inputMessage.c_str());

    for(int i = 0; i < playerList.getSize(); i++){
      Serial.println(i);
      currentPlayer = playerList.get(i);
      Serial.println(currentPlayer.name);
      if (request->hasParam(currentPlayer.name)) {
        all = false;    
        Serial.println("to on");
        memcpy(playerAddressX,currentPlayer.macAddr, 6);
        esp_now_send(playerAddressX, (uint8_t *) &outgoingSetpoints, sizeof(outgoingSetpoints));
      }
    }

    if(all){
      Serial.println("to all");
      esp_now_send(NULL, (uint8_t *) &outgoingSetpoints, sizeof(outgoingSetpoints));
    }
    
    Serial.println((String)inputMessage);
    request->send(LittleFS, "/index.html", String(), false);
  });

  // start server
  server.begin();
}


void loop() {
  static unsigned long lastEventTime = millis();
  static const unsigned long EVENT_INTERVAL_MS = 5000;
  if ((millis() - lastEventTime) > EVENT_INTERVAL_MS) {
    events.send("ping", NULL, millis());
    lastEventTime = millis();

  }
}